# -*- coding: utf-8 -*-


from __future__ import print_function
import networkx as nx
from networkx.readwrite import json_graph
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import json


class Ortho_Network:
    """Calculates and holds data about interactions in form of network,
    exportable to other software (e.g. Cytoscape) or drawable by matplotlib.

    Attribs:
        inter_df (pandas.DataFrame): DataFrame containing genetic interactions
        nwrk (networkx.Graph): network created upon Ortho_Network.inter_df
    """
    def __init__(self,
                 inter_df):
        self.inter_df = inter_df
        self.nwrk = None
        self.sub_nwrk = None

    def create_nwrk(self,
                    nodes_cols,
                    attribs_cols):
        """Return Ortho_Network.nwrk upon pandas.DataFrame.

        Parameters
        -------
        nodes_cols: list
            Columns to take as nodes.
        attribs_cols: list
            Columns to take as attributes.

        Returns
        -------
        P_CRAWLER.Ortho_Network.nwrk
            Interactions-based network. networkx.classes.graph.Graph derivate.
        """
        self.nwrk = nx.from_pandas_dataframe(self.inter_df,
                                             nodes_cols[0],
                                             nodes_cols[1],
                                             attribs_cols)

    def get_subgrps(self):
        self.sub_nwrk = [i for i in nx.connected_component_subgraphs(self.nwrk)]

    def write_nwrk(self,
                   out_file_name,
                   out_file_format):
        """Write Ortho_Network.nwrk to file readable to other software.

        Args:
            out_file_name (str): file name to save as
            out_file_format (str): file format to save as. Available formats are:
            graphml, gefx, gml, json
        """
        if out_file_format.lower() == "graphml":
            nx.write_graphml(self.nwrk, out_file_name)
        elif out_file_format.lower() == "gefx":
            nx.write_gexf(self.nwrk, out_file_name)
        elif out_file_format.lower() == "gml":
            nx.write_gml(self.nwrk, out_file_name)
        elif out_file_format.lower() == "json":
            with open("{0}.{1}".format(out_file_name, out_file_format), "w") as fout:
                fout.write(json.dumps(json_graph.node_link_data(self.nwrk)))

    def draw_nwrk(self,
                  width=20,
                  height=20,
                  dpi=None,
                  node_size=5,
                  save_2_file=False,
                  out_file_name="network.png",
                  sub_nwrk=False):
        """Return matplotlib.pyplot.figure of Ortho_Network.nwrk and/or write it to
        <*.png> file.

        Args:
            width (int): figure width in inches. Set as speciefied in
            matplotlibrc file when <None>. Default: <20>
            height (int): figure height in inches. Set as speciefied in
            matplotlibrc file when <None>. Default: <20>
            dpi (int): figure resolution. Set as speciefied in
            matplotlibrc file when <None>. Default: <None>
            node_size (int): size of the nodes. Default: <5>
            save_2_file (bool): write to file when <True>. Default: <False>
            out_file_name (str): file name to save as

        """
        if sub_nwrk is True:
            nwrk = self.sub_nwrk
        else:
            nwrk = self.nwrk
        plt.figure(figsize=(width, height))
        nx.draw_networkx(nwrk,
                         node_size=node_size,
                         node_color="r",
                         node_alpha=0.4,
                         with_labels=False)
        if save_2_file is True:
            plt.savefig(out_file_name,
                        dpi=dpi)
        else:
            pass



# -*- coding: utf-8 -*-


from __future__ import print_function
from prwlr.errors import *
import numpy as np
import pandas as pd
from scipy.spatial import distance as dist


class Profile:
    """
    Profile object.
    """
    _positive_sign = "+"
    _negative_sign = "-"

    def __init__(self,
                 reference,
                 query):
        try:
            iter(reference)
        except TypeError:
            raise ProfileConstructorError("Reference must be an iterable.")
        try:
            iter(query)
        except TypeError:
            raise ProfileConstructorError("Query must be an iterable.")
        self.reference = set(reference)
        self.query = sorted(tuple(set(query)))
        self._construct()

    def __len__(self):
        return len(self.profile)

    def __repr__(self):
        return self.to_string()

    def __lt__(self,
               other):
        return self.profile < other.profile

    def __le__(self,
               other):
        return self.profile <= other.profile

    def __gt__(self,
               other):
        return self.profile > other.profile

    def __ge__(self,
               other):
        return self.profile >= other.profile

    def __ne__(self,
               other):
        return self.profile != other.profile

    def __eq__(self,
               other):
        return self.profile == other.profile and self.query == other.query

    def __hash__(self):
        return hash(self.profile)

    def _construct(self):
        """
        Construct profile from Profile.reference and Profile.query.
        """
        self.profile = tuple(True if i in self.reference else False for i in self.query)

    def _bind(self):
        """
        Return zipped Profile.query and Profile.profile.
        """
        return tuple(zip(self.query, self.profile))

    def _convert(self,
                 positive_sign,
                 negative_sign):
        """
        Convert profile to given sign.
        """
        return tuple(positive_sign if i is True else negative_sign for i in self.profile)

    def isall(self,
              queries):
        """
        Returns <True> if all positions from queries are present in the
        profile.
        """
        if all([dict(tuple(zip(self.query, self.profile)))[i] for i in queries]):
            return True
        else:
            return False

    def isany(self,
              queries):
        """
        Return <True> if any of the positions from queries is present in the
        profile.
        """
        if any([dict(tuple(zip(self.query, self.profile)))[i] for i in queries]):
            return True
        else:
            return False

    def to_string(self,
                  positive_sign=_positive_sign,
                  negative_sign=_negative_sign):
        """
        Return profile as str.
        """
        if positive_sign is not None and negative_sign is not None:
            return "".join(self._convert(positive_sign,
                                         negative_sign))
        else:
            return str(self.profile)

    def to_list(self,
                positive_sign="+",
                negative_sign="-"):
        """
        Retrun profile as list.
        """
        if positive_sign is not None and negative_sign is not None:
            return list(self._convert(positive_sign,
                                      negative_sign))
        else:
            list(self.profile)

    def to_tuple(self,
                 positive_sign="+",
                 negative_sign="-"):
        """
        Return profile as tuple.
        """
        if positive_sign is not None and negative_sign is not None:
            return tuple(self._convert(positive_sign,
                                       negative_sign))
        else:
            return tuple(self.profile)

    def to_array(self,
                 positive_sign="+",
                 negative_sign="-"):
        """
        Return profile as an numpy.array.
        """
        if positive_sign is not None and negative_sign is not None:
            return np.array(self._convert(positive_sign,
                                          negative_sign))
        else:
            return np.array(self.profile)

    def to_series(self,
                  positive_sign="+",
                  negative_sign="-"):
        """
        Return profile as pandas.Series
        """
        if positive_sign is not None and negative_sign is not None:
            return pd.Series(self._convert(positive_sign,
                                           negative_sign))
        else:
            return pd.Series(self.profile)

    def calculate_pss(self,
                      profile,
                      ignore=None,
                      method="pairwise"):
        """
        Calculate Profiles Similarity Score.
        """
        if len(self) != len(profile):
            raise ProfileError("Different profiles' lengths")
        prof_1 = self
        prof_2 = profile
        if ignore:
            for i in ignore:
                try:
                    prof_1.profile = list(prof_1.profile)
                    del prof_1.profile[prof_1.query.index(i)]
                    prof_1.profile = tuple(prof_1.profile)
                except IndexError:
                    raise ProfileError("Element to ignore not in profile")
                try:
                    prof_2.profile = list(prof_2.profile)
                    del prof_2.profile[prof_2.query.index(i)]
                    prof_2.profile = tuple(prof_2.profile)
                except IndexError:
                    raise ProfileError("Element to ignore not in profile")
        if method == "pairwise":
            return sum(a == b for a, b in zip(prof_1.profile, prof_2.profile))
        elif method == "jaccard":
            return dist.jaccard(prof_1.profile, prof_2.profile)
        elif method == "yule":
            return dist.yule(prof_1.profile, prof_2.profile)
        elif method == "dice":
            return dist.dice(prof_1.profile, prof_2.profile)
        elif method == "hamming":
            return dist.hamming(prof_1.profile, prof_2.profile)
        elif method == "kulsinski":
            return dist.kulsinski(prof_1.profile, prof_2.profile)
        elif method == "rogerstanimoto":
            return dist.rogerstanimoto(prof_1.profile, prof_2.profile)
        elif method == "russellrao":
            return dist.russellrao(prof_1.profile, prof_2.profile)
        elif method == "sokalmichener":
            return dist.sokalmichener(prof_1.profile, prof_2.profile)

    def get_present(self):
        """
        Return elements present in the profile.
        """
        return tuple(k for k, v in self._bind() if v is True)

    def get_absent(self):
        """
        Return element absent in the profile.
        """
        return tuple(k for k, v in self._bind() if v is False)



__version__ = '0.0.1'
__author__ = 'Dariusz Izak IBB PAS'


__all__ = [
    "apis",
    "databases",
    "errors",
    "network",
    "profiles",
    "stats",
    "utils"
]

try:
    import pandas as _pd
    import numpy as _np
    import subprocess as _sp
    from prwlr.core import *
except ImportError:
    print('Could not import dependencies. Ignore if running setup.')



# -*- coding: utf-8 -*-


from __future__ import print_function
import requests as rq
import pandas as pd
from tqdm import tqdm
import pathos.threading as ptth


class Columns:
    """
    Container for the columns names defined in this module.
    """
    GENOME_ID = "GENOME_ID"
    NAMES = "NAMES"
    DESCRIPTION = "DESCRIPTION"
    KEGG_ORG_ID = "KEGG_ORG_ID"
    NAME = "NAME"
    TAXON_ID = "TAXON_ID"
    KEGG_ID = "KEGG_ID"
    ORF_ID = "ORF_ID"
    ORG_GENE_ID = "ORG_GENE_ID"
    dtypes = {TAXON_ID: "uint32"}


class KEGG_API(Columns):
    """Provides connectivity with the KEGG database. Functions ending with <tbl>
    download files provided by KEGG but DO NOT modify them. Modifications
    needed for data processing are made on pandas.DataFrame.
    """
    def __init__(self):
        self.home = "http://rest.kegg.jp"
        self.operations = {"db_statistics": "info",
                           "list_entry_ids": "list",
                           "find_by_keyword": "find",
                           "get_by_entry_no": "get",
                           "conv_2_outside_ids": "conv",
                           "find_X_ref": "link"}
        self.databases = {"pathway": "path",
                          "brite": "br",
                          "module": "md",
                          "orthology": "ko",
                          "genome": "genome",
                          "genomes": "gn",
                          "genes": "genes",
                          "enzyme": "ec"}
        self.organisms_ids_df = None
        self.id_conversions = {"ncbi_gene": "ncbi-geneid",
                               "ncbi_prot": "ncbi-proteinid",
                               "uniprot": "uniprot",
                               "kegg_id": "genes"}
        self.id_conversions_df = None
        self.org_db_X_ref_df = None
        self.query_ids_found = []
        self.query_ids_not_found = []

    def get_organisms_ids(self,
                          out_file_name,
                          skip_dwnld=False):
        """Get KEGG's organisms' IDs, genomes IDs and definitions. Data are
        downloaded to a local file and then made into pandas.DataFrame. File
        can be reused. Necessary for KEGG_API.org_name_2_kegg_id.

        Args:
            out_file_name (str): name for file to be downloaded
            skip_dwnld (bool): read existing file when <True>. Default <False>
        """
        if skip_dwnld is True:
            pass
        else:
            url = "{0}/{1}/{2}".format(self.home,
                                       self.operations["list_entry_ids"],
                                       self.databases["genome"])
            res = rq.get(url)
            with open(out_file_name, "wb") as fout:
                fout.write(res.content)
        self.organisms_ids_df = pd.read_csv(out_file_name,
                                            names=[self.GENOME_ID,
                                                   self.NAMES,
                                                   self.DESCRIPTION],
                                            header=None,
                                            sep="\t|;",
                                            engine="python",
                                            error_bad_lines=False,
                                            warn_bad_lines=True)
        temp_sub_df = self.organisms_ids_df[self.NAMES].str.split(",", expand=True)
        temp_sub_df.columns = [self.KEGG_ORG_ID, self.NAME, self.TAXON_ID]
        self.organisms_ids_df.drop(self.NAMES, axis=1, inplace=True)
        self.organisms_ids_df = pd.concat([self.organisms_ids_df, temp_sub_df], axis=1)
        self.organisms_ids_df.replace({"genome:": ""},
                                      regex=True,
                                      inplace=True)
        self.organisms_ids_df.dropna(inplace=True)
        self.organisms_ids_df = self.organisms_ids_df.astype({k: v for k, v in self.dtypes.items()
                                                              if k in self.organisms_ids_df.columns})

    def org_name_2_kegg_id(self,
                           organism,
                           assume_1st=True):
        """Return KEGG's organisms' IDs (str) when queried  with a regular
        (natural) biological name. Case-sensitive. Uses KEGG_API.organisms_ids_df
        generated by KEGG_API.get_organisms_ids. Necessary for creation of ids
        list which is then passed to Genome.KO_list_profiler.

        Args:
            organism (str): biological organism's name to query against
            the KEGG's IDs
            assume_1st (bool): return the first item if more than one hit when
            <True> (default)
        """
        org_bool = self.organisms_ids_df[self.DESCRIPTION].str.contains(organism)
        organism_ser = self.organisms_ids_df[org_bool]
        if len(organism_ser) == 0:
            print("No record found for {}".format(organism))
            self.query_ids_not_found.append(organism)
        elif len(organism_ser) > 1:
            print("More than one record for this query\n{}".format(organism_ser[[self.DESCRIPTION,
                                                                                 self.KEGG_ORG_ID]]))
            if assume_1st is True:
                self.query_ids_found.append(organism)
                return organism_ser[self.KEGG_ORG_ID].iloc[0]
            else:
                return None
        else:
            self.query_ids_found.append(organism)
            return organism_ser[self.KEGG_ORG_ID].iloc[0]

    def get_org_db_X_ref(self,
                         organism,
                         target_db,
                         out_file_name,
                         skip_dwnld=False,
                         strip_prefix=True,
                         drop_duplicates=True):
        """Get desired KEGG's database entries linked with all the genes from
        given organism. Data are downloaded to a local file and then made into
        pandas.DataFrame. File can be reused. Necessary for
        KEGG_API.get_ortho_db_entries and Ortho_Interactions.KO_based_appender.

        Args:
            organism (str): organism name. Provide whitespace-separated full
            species name. Uses pandas.series.str.contains method.
            targed_db (str): dict key for KEGG_API.databases of desired
            database.
            out_file_name (str): name for file to be downloaded
            skip_dwnld (bool) = read existing file when <True>. Default <False>
        """
        org_id = self.org_name_2_kegg_id(organism)
        if skip_dwnld is True:
            pass
        else:
            url = "{0}/{1}/{2}/{3}".format(self.home,
                                           self.operations["find_X_ref"],
                                           self.databases[target_db],
                                           org_id)
            res = rq.get(url)
            with open(out_file_name, "wb") as fout:
                fout.write(res.content)
        self.org_db_X_ref_df = pd.read_csv(out_file_name,
                                           names=[self.ORF_ID, self.KEGG_ID],
                                           header=None,
                                           sep="\t")
        if strip_prefix is True:
            self.org_db_X_ref_df.replace({"{}:".format(org_id): "",
                                          "{}:".format(self.databases[target_db]): ""},
                                         regex=True,
                                         inplace=True)
        if drop_duplicates:
            self.org_db_X_ref_df.drop_duplicates(
                subset=[self.ORF_ID],
                keep=False,
                inplace=True,
            )
            self.org_db_X_ref_df.drop_duplicates(
                subset=[self.KEGG_ID],
                keep=False,
                inplace=True,
            )

    def get_KOs_db_X_ref(self,
                         filename,
                         target_db='genes',
                         skip_dwnld=False,
                         strip_prefix=True,
                         squeeze=True,
                         sep="\t",
                         threads=1):
        """
        Get desired KEGG's database entries linked with KEGG Orthology Group.
        Data are downloaded to a local file and then made into pandas.DataFrame.
        File can be reused.

        Parameters
        -------
        filename: str
            Name of the file to download.
        targed_db: str
            Key for KEGG_API.databases of desired database.
        skip_dwnld: bool, default <False>
            Read already downloaded file if <True>
        squeeze: bool, default <True>
            Compress list of the organisms to list for each KEGG Orthology
            Group
        sep: str, default: <\t>
            Delimiter to use.
        threads: int, default <1>
            Number of threads to spawn during download.

        Uses
        -------
        KEGG_API.org_db_X_ref_df: pandas.DataFrame
            Source of the KEGG Orthology Groups.

        Sets
        -------
        KEGG_API.KOs_db_X_ref_df: pandas.DataFrame
            DataFrame of KEGG Orthology Group ID and belonging organisms and
            genes
        """
        def f(i):
            print("{} ".format(i), flush=True, end='\r')
            res = rq.get('{}/{}/{}/{}'.format(
            self.home,
            self.operations['find_X_ref'],
            self.databases[target_db],
            i,
            ))
            with open(filename, 'ab') as fout:
                fout.write(res.content)
        if not skip_dwnld:
            if threads > 1:
                ptth.ThreadPool(threads).map(
                    f,
                    self.org_db_X_ref_df[self.KEGG_ID]
                )
            else:
                map(f, self.org_db_X_ref_df[self.KEGG_ID])
        self.KOs_db_X_ref_df = pd.read_csv(filename,
                                           names=[self.KEGG_ID,
                                                  self.ORG_GENE_ID],
                                           header=None,
                                           sep=sep)
        if strip_prefix:
            self.KOs_db_X_ref_df.replace({"{}:".format(self.databases["orthology"]): ""},
                                         regex=True,
                                         inplace=True)
            self.KOs_db_X_ref_df.replace({":.+": ""},
                                         regex=True,
                                         inplace=True)
        if squeeze:
            self.KOs_db_X_ref_df = self.KOs_db_X_ref_df.groupby(
                by=[self.KEGG_ID]
                )[self.ORG_GENE_ID].apply(list).to_frame().reset_index()

    def get_db_entries(self,
                       out_file_name):
        """Get full database by quering entries from
        KEGG_API.org_db_X_ref_df and download them into a local file.
        Necessary for Genome.parse_KO_db. The only func that does NOT convert
        downloaded file into pandas.DataFrame. Uses KEGG_API.get_db_X_ref_df.

        Args:
            out_file_name (str): name for file to be downloaded
        """
        entries = self.org_db_X_ref_df[self.KEGG_ID].drop_duplicates()
        for i in tqdm(entries):
            url = "{0}/{1}/{2}".format(self.home,
                                       self.operations["get_by_entry_no"],
                                       i)
            res = rq.get(url)
            with open(out_file_name, "ab") as fout:
                fout.write(res.content)


class CostanzoAPI:
    """Provides connectivity with the Costanzo's SOM website of the Genetic
    Landscape of the Cell project, allowing data files download.

    Attribs:
        home (str): Costanzo's SOM home page address
        raw (str): raw data link and file name
        raw_matrix (str): raw data genetic interactions matrix link and file
        name, Java Treeview format
        lenient_cutoff (str): GIS_P < 0.05 cutoff link and file name
        intermediate_cutoff (str): |genetic interaction score| > 0.08,
        GIS_P < 0.05 cutoff link and file name
        stringent_cutoff (str): genetic interaction score < -0.12,
        GIS_P < 0.05 or genetic interaction score > 0.16, GIS_P < 0.05 link
        and file name
        bioprocesses (str): bioprocesses annotations
        chemical_genomics (str): chemical genomics data
        query_list (str): query ORFs list
        array_list (str): array ORFs list
    """

    def __init__(self):
        self.home = {"v1": "http://drygin.ccbr.utoronto.ca/~costanzo2009",
                     "v2": "http://thecellmap.org/costanzo2016/"}
        self.data = {"v1": {"raw": "sgadata_costanzo2009_rawdata_101120.txt.gz",
                            "raw_matrix": "sgadata_costanzo2009_rawdata_matrix_101120.txt.gz",
                            "lenient_cutoff": "sgadata_costanzo2009_lenientCutoff_101120.txt.gz",
                            "intermediate_cutoff": "sgadata_costanzo2009_intermediateCutoff_101120.txt.gz",
                            "stringent_cutoff": "sgadata_costanzo2009_stringentCutoff_101120.txt.gz",
                            "bioprocesses": "bioprocess_annotations_costanzo2009.xls",
                            "chemical_genomics": "chemgenomic_data_costanzo2009.xls",
                            "query_list": "sgadata_costanzo2009_query_list_101120.txt",
                            "array_list": "sgadata_costanzo2009_array_list.txt"},
                       "v2": {"pairwise": "data_files/Raw%20genetic%20interaction%20datasets:%20Pair-wise%20interaction%20format.zip",
                              "matrix": "data_files/Raw%20genetic%20interaction%20datasets:%20Matrix%20format.zip",
                              "interaction_profile_similarity_matrices": "data_files/Genetic%20interaction%20profile%20similarity%20matrices.zip"}}

    def get_data(self,
                 data,
                 output_directory=".",
                 sga_version="v2"):
        """Get files from Costanzo's SOM website.

        Args:
            data (str): specifies the file to be downloaded.
            <raw> for raw dataset,
            <raw_matrix> for raw genetic interactions matrix,
            <lenient_cutoff> for lenient dataset,
            <intermediate_cutoff> for intermediate dataset,
            <stringent_cutoff> for stringent dataset,
            <bioprocesses> for bioprocesses dataset,
            <chemical_genomics> for chemical genomics dataset,
            <query_list> for list of query ORFs names,
            <array_list> for list of array ORFs names
            out_file_name (str): name for file to be downloaded. Automatically
            same as appropriate Costanzo_API attrib when set to <None>
        """
        if data not in list(self.data[sga_version].keys()):
            raise ValueError("unknown option for data arg")
        url = "{0}/{1}".format(self.home[sga_version],
                               self.data[sga_version][data])
        out_file_name = self.data[sga_version][data].replace("data_files/", "").replace("%20", "_").replace(":", "-")
        res = rq.get(url)
        with open("{}/{}".format(output_directory, out_file_name), "wb") as fout:
            fout.write(res.content)



# -*- coding: utf-8 -*-


from __future__ import print_function
import re
import pathos.multiprocessing as ptmp
import numpy as np
import pandas as pd
import tempfile
from prwlr.apis import KEGG_API as _KEGG_API
from prwlr.apis import Columns as _ApisColumns
from prwlr.errors import *
from prwlr.profiles import Profile as _Profile
from prwlr.utils import *


class Columns(_ApisColumns):
    """
    Container for the columns names defined in this module.
    """
    # Query and Array suffixes for query and array discrimination.
    QUERY_SUF = "_Q"
    ARRAY_SUF = "_A"
    # apis column names suffixed.
    KEGG_ID_A = "{}{}".format(_ApisColumns.KEGG_ID, ARRAY_SUF)
    KEGG_ID_Q = "{}{}".format(_ApisColumns.KEGG_ID, QUERY_SUF)
    ORF_ID_A = "{}{}".format(_ApisColumns.ORF_ID, ARRAY_SUF)
    ORF_ID_Q = "{}{}".format(_ApisColumns.ORF_ID, QUERY_SUF)
    # SGAs column names
    ORF_A = "ORF{}".format(ARRAY_SUF)
    GENE_A = "GENE{}".format(ARRAY_SUF)
    SMF_A = "SMF{}".format(ARRAY_SUF)
    SMF_SD_A = "SMF_SD{}".format(ARRAY_SUF)
    ORF_Q = "ORF_Q".format(QUERY_SUF)
    GENE_Q = "GENE_Q".format(QUERY_SUF)
    SMF_Q = "SMF_Q".format(QUERY_SUF)
    SMF_SD_Q = "SMF_SD_Q".format(QUERY_SUF)
    DMF = "DMF"
    DMF_SD = "DMF_SD"
    GIS = "GIS"
    GIS_SD = "GIS_SD"
    GIS_P = "GIS_P"
    STR_ID_Q = "STR_ID_Q".format(QUERY_SUF)
    GENE_Q = "GENE_Q".format(QUERY_SUF)
    STR_ID_A = "STR_ID{}".format(ARRAY_SUF)
    TEMP = "TEMP"
    # Bioprocesses and permutation internal dataframe column names.
    ORF = "ORF"
    GENE = "GENE"
    BIOPROC = "BIOPROC"
    # KEGG and ProfInt and permutation internal dataframe column names
    AUTH = "AUTH"
    DEF = "DEF"
    ENTRY = "ENTRY"
    GENES = "GENES"
    JOURN = "JOURN"
    NAME = "NAME"
    ORGS = "ORGS"
    ORGS_A = "ORGS{}".format(ARRAY_SUF)
    ORGS_Q = "ORGS{}".format(QUERY_SUF)
    PROF = "PROF"
    REF = "REF"
    SEQ = "SEQ"
    TITLE = "TITLE"
    PSS = "PSS"
    PROF_Q = "PROF_Q".format(QUERY_SUF)
    PROF_A = "PROF{}".format(ARRAY_SUF)
    dtypes = {GIS: "float32",
              GIS_SD: "float32",
              SMF_Q: "float32",
              SMF_A: "float32",
              DMF: "float32",
              DMF_SD: "float32",
              PSS: "uint8"}


class KEGG(Columns):
    """
    Parses data downloaded with prwlr.apis and restructures them.

    Parameters
    -------
    listed: list of dicts
        Data from parsed KEGG database.
    """
    def __init__(self,
                 database_type):
        self.database_type = database_type.lower()
        self._api = _KEGG_API()

    def parse_database(self,
                       filename,
                       cleanup=True,
                       remove_from_orgs=None):
        """Return KEGG.listed (list of dicts) which contains information from
        the file downloaded by KEGG_API.get_ortho_db_entries.

        Args:
            filename (str): file name to parse
        """
        with open(filename, "r") as fin:
            file_str = fin.read()
            entries_list = file_str.split("///")
            if len(entries_list) < 2:
                raise ParserError("No split sign. Check if <///> in file.")

        def f(i):
            entry_dict = {}
            entry = re.findall("ENTRY.+", i)
            if len(entry) > 0:
                entry_dict[self.ENTRY] = entry[0].replace("ENTRY", "").replace("KO", "").strip()
            name = re.findall("NAME.+", i)
            if len(name) > 0:
                entry_dict[self.NAME] = name[0].replace("NAME", "").strip()
            definition = re.findall("DEFINITION.+", i)
            if len(definition):
                entry_dict[self.DEF] = definition[0].replace("DEFINITION", "").strip()
            reference = re.findall("REFERENCE.+", i)
            if len(reference) > 0:
                entry_dict[self.REF] = reference[0].replace("REFERENCE", "").strip()
            authors = re.findall("AUTHORS.+", i)
            if len(authors) > 0:
                entry_dict[self.AUTH] = authors[0].replace("AUTHORS", "").strip()
            title = re.findall("TITLE.+", i)
            if len(title) > 0:
                entry_dict[self.TITLE] = title[0].replace("TITLE", "").strip()
            journal = re.findall("JOURNAL.+", i)
            if len(journal) > 0:
                entry_dict[self.JOURN] = journal[0].replace("JOURNAL", "").strip()
            sequence = re.findall("SEQUENCE.+", i)
            if len(sequence) > 0:
                entry_dict[self.SEQ] = sequence[0].replace("SEQUENCE", "").replace("[", "").replace("]", "").strip()
            genes_blk_comp = re.compile("GENES.+\n^\s+\w{3}:\s.+^\w", re.DOTALL | re.MULTILINE)
            genes_blk_list = genes_blk_comp.findall(i)
            re.purge()
            if len(genes_blk_list) > 0:
                genes_blk_str = genes_blk_list[0]
                orgs_n_genes = re.findall("\w{3}:.+", genes_blk_str)
                orgs = []
                genes = []
                for i in orgs_n_genes:
                    if ": " in i:
                        orgs.append(i.split(": ")[0])
                        genes.append(i.split(": ")[1])
                    else:
                        orgs.append(i)
                entry_dict[self.GENES] = genes
                entry_dict[self.ORGS] = orgs
            return entry_dict
        listed = ptmp.ProcessingPool().map(f, entries_list)
        df = pd.DataFrame(listed)
        if cleanup is True:
            df = df.drop_duplicates(subset=[self.ENTRY],
                                    keep="first")
            df.index = list(range(len(df)))
            df.dropna(how="all",
                      inplace=True)
            df.dropna(subset=[self.ENTRY,
                              self.ORGS],
                      inplace=True)
        if remove_from_orgs is not None:
            for i in remove_from_orgs:
                df[self.ORGS] = df[self.ORGS].apply(lambda x: remove_from_list(i, x))
        self.database = df

    def parse_organism_info(self,
                            organism,
                            reference_species,
                            IDs=None,
                            X_ref=None,
                            KOs=None,
                            strip_prefix=True,
                            IDs_only=False,
                            threads=6):
        print("Getting the organisms' KEGG IDs...")
        if IDs:
            self._api.get_organisms_ids(IDs, skip_dwnld=True)
        else:
            IDs_tmp = tempfile.NamedTemporaryFile(delete=True)
            self._api.get_organisms_ids(IDs_tmp.name, skip_dwnld=False)
            IDs_tmp.close()
        self.reference_species = [self._api.org_name_2_kegg_id(i) for i in reference_species
                                  if i not in self._api.query_ids_not_found]
        self.reference_species = [i.upper() for i in self.reference_species if i is not None]
        self.name_ID = dict(list(zip([i for i in reference_species
                                 if i not in self._api.query_ids_not_found],
                                self.reference_species)))
        self.ID_name = dict(list(zip(self.reference_species, [i for i in reference_species
                                                         if i not in self._api.query_ids_not_found])))
        if IDs_only:
            return
        print("Getting the ORF-Orthology Group Cross Reference...")
        if X_ref:
            self._api.get_org_db_X_ref(organism=organism,
                                       target_db=self.database_type,
                                       out_file_name=X_ref,
                                       skip_dwnld=True,
                                       strip_prefix=True)
        else:
            X_ref_tmp = tempfile.NamedTemporaryFile(delete=True)
            self._api.get_org_db_X_ref(organism=organism,
                                       target_db=self.database_type,
                                       out_file_name=X_ref_tmp.name,
                                       skip_dwnld=False,
                                       strip_prefix=True)
            X_ref_tmp.close()
        self.X_reference = self._api.org_db_X_ref_df
        print("Getting the Organisms List for Each of The Orthology Group...")
        if KOs:
            self._api.get_KOs_db_X_ref(filename=KOs,
                                       skip_dwnld=True)
        else:
            KOs_temp = tempfile.NamedTemporaryFile(delete=True)
            self._api.get_KOs_db_X_ref(filename=KOs_temp.name,
                                       skip_dwnld=False,
                                       squeeze=True,
                                       threads=threads)
            KOs_temp.close()
        try:
            pd.testing.assert_series_equal(
                self.X_reference[self.KEGG_ID].
                drop_duplicates().
                sort_values().
                reset_index(drop=True),
                self._api.KOs_db_X_ref_df[self.KEGG_ID].
                drop_duplicates().
                sort_values().
                reset_index(drop=True)
            )
            self.KO_organisms = self._api.KOs_db_X_ref_df
        except AssertionError:
            if threads > 1:
                raise ParserError(
                    """{} of X_reference and KO_organisms are different. This
                    might be caused by the server access denial. Try
                    deacreasing number of threads""".format(
                        self.KEGG_ID
                    )
                )
            else:
                raise ParserError(
                    """{} of X_reference and KO_organisms are different.""".
                    format(self.KEGG_ID)
                )
        self.organism_info = pd.merge(
            left=self.X_reference,
            right=self.KO_organisms,
            on=self.KEGG_ID,
        )
        self.organism_info[self.PROF] = self.organism_info[self.ORG_GENE_ID].apply(
            lambda x: _Profile(
                x,
                [i.lower() for i in self.name_ID.values()],
            )
        )
        self.organism_info.drop(
            columns=self.ORG_GENE_ID,
            inplace=True,
        )
        self.organism_info.drop_duplicates(inplace=True)



class SGA1(Columns):
    """
    Port from interactions.Ortho_Interactions. Meant to work just with SGA v1.

    Notes
    -------

    """
    def __init__(self):
        self.names = (('Query_ORF', self.ORF_Q),
                      ('Query_gene_name', self.GENE_Q),
                      ('Array_ORF', self.ORF_A),
                      ('Array_gene_name', self.GENE_A),
                      ('Genetic_interaction_score', self.GIS),
                      ('Standard_deviation', self.GIS_SD),
                      ('p-value', self.GIS_P),
                      ('Query_SMF', self.SMF_Q),
                      ('Query_SMF_standard_deviation', self.SMF_SD_Q),
                      ('Array_SMF', self.SMF_A),
                      ('Array_SMF_standard_deviation', self.SMF_SD_A),
                      ('DMF', self.DMF),
                      ('DMF_standard_deviation', self.DMF_SD))

    def parse(self,
              filename,
              remove_white_spaces=True,
              in_sep="\t",
              cleanup=True):
        """Return Ortho_Interactions.interact_df (pandas.DataFrame) from
        parsed <csv> file. The minimal filtration is based of a given GIS_P
        and presence of DMF value. Further filtration results in DMF
        higher/lower than both SMFs.

        Args:
            sga (str): name of file to parse
            sga_ver (int) = costanzo dataframe version
            excel (bool): pandas.read_excel when <True>. pandas.read_csv when
            <False> (default).
            p_value (float): maximum GIS_P for filtering
            DMF_type (str): positive -> DMF > both SMFs
                            negative -> DMF < both SMFs
                            neutral  -> DMF not <None> (default)
                            raw      -> no filter
            remove_white_spaces (bool): replaces whitespaces from col names
            with <_> when True (default)
            in_sep (str): separator for pandas.read_csv method
        """
        self.sga = pd.read_csv(filename,
                               sep=in_sep,
                               names=[k for k, v in self.names],
                               error_bad_lines=False,
                               warn_bad_lines=True)
        if remove_white_spaces is True:
            self.sga.columns = [i.replace(" ", "_") for i in self.sga.columns]
        self.sga.rename(columns=dict(self.names), inplace=True)
        self.sga = self.sga.astype({k: v for k, v in self.dtypes.items()
                                    if k in self.sga.columns})
        if cleanup:
            self.sga = self.sga.dropna().drop_duplicates().reset_index(drop=True)


class SGA2(Columns):
    """
    Port from interactions.Ortho_Interactions. Meant to work just with SGA v2.

    Notes
    -------

    """
    def __init__(self):
        self.names = {"Query_Strain_ID": self.STR_ID_Q,
                      "Query_allele_name": self.GENE_Q,
                      "Array_Strain_ID": self.STR_ID_A,
                      "Array_allele_name": self.GENE_A,
                      "Arraytype/Temp": self.TEMP,
                      "Genetic_interaction_score_(ε)": self.GIS,
                      "P-value": self.GIS_P,
                      "Query_single_mutant_fitness_(SMF)": self.SMF_Q,
                      "Array_SMF": self.SMF_A,
                      "Double_mutant_fitness": self.DMF,
                      "Double_mutant_fitness_standard_deviation": self.DMF_SD}

    def parse(self,
              filename,
              remove_white_spaces=True,
              in_sep="\t"):
        """Return Ortho_Interactions.interact_df (pandas.DataFrame) from
        parsed <csv> file. The minimal filtration is based of a given GIS_P
        and presence of DMF value. Further filtration results in DMF
        higher/lower than both SMFs.

        Args:
            filename (str): name of file to parse
            p_value (float): maximum GIS_P for filtering
            DMF_type (str): positive -> DMF > both SMFs
                            negative -> DMF < both SMFs
                            neutral  -> DMF not <None> (default)
                            raw      -> no filter
            remove_white_spaces (bool): replaces whitespaces from col names
            with <_> when True (default)
            in_sep (str): separator for pandas.read_csv method
        """
        self.sga = pd.read_csv(filename, sep=in_sep)
        if remove_white_spaces is True:
            self.sga.columns = [i.replace(" ", "_") for i in self.sga.columns]
        self.sga.rename(columns=self.names, inplace=True)
        ORF_Q_col = self.sga[self.STR_ID_Q].str.split("_", expand=True)[0]
        ORF_A_col = self.sga[self.STR_ID_A].str.split("_", expand=True)[0]
        ORF_Q_col.name = self.ORF_Q
        ORF_A_col.name = self.ORF_A
        self.sga = pd.concat([ORF_Q_col, ORF_A_col, self.sga], axis=1)
        self.sga = self.sga.astype({k: v for k, v in self.dtypes.items()
                                    if k in self.sga.columns})


class AnyNetwork(Columns):
    """
    Parses and holds data of any type of network.
    """
    def __init__(self):
        pass

    def parse(self,
              filename,
              sep="\t",
              excel=False,
              sheet_name=None,
              ORF_query_col=None,
              ORF_array_col=None,
              **kwargs):
        """
        Parses network csv file. Checks whether columns names in the csv file
        correspond with the databases.Columns.
        """
        if ORF_query_col is None or ORF_array_col is None:
            raise ParserError("No ORF_query or ORF_array column name.")
        if excel is True:
            self.sga = pd.read_excel(filename, sheet_name=sheet_name)
        else:
            self.sga = pd.read_csv(filename, sep=sep)
        self.sga.rename(columns={ORF_query_col: self.ORF_Q,
                                 ORF_array_col: self.ORF_A},
                        inplace=True)
        if len(kwargs) > 0:
            self.sga.rename(columns=kwargs,
                            inplace=True)


class Bioprocesses(Columns):
    """
    Port from interactions.Ortho_Interactions. Meant to work with
    bioprocesses_annotations.costanzo2009.
    """

    def __init__(self):
        self.names = [self.ORF,
                      self.GENE,
                      self.BIOPROC]

    def parse(self,
              filename):
        """Return Ortho_Interactions.bio_proc_df (pandas.DataFrame) from parsed
        <csv> or <xls> file.

        Parameters
        -------
        filename: str
            Name of file to parse.
        """
        self.bioprocesses = pd.read_excel(filename,
                                          names=self.names)
        self.bioprocesses = self.bioprocesses.astype({k: v for k, v in self.dtypes.items()
                                                     if k in self.bioprocesses.columns})



# -*- coding: utf-8 -*-


from __future__ import print_function
import math


def isiniterable(query_iterable,
                 reference_iterable,
                 all_present=False):
    """
    Returns <True> or <False> depending on whether elements from one iterable
    are found in the second.

    Parameters
    -------
    query_iterable: list, tuple, set
        Iterable from which the elements presence is checked in reference_iterable.
    reference_iterable: list, tuple, set
        Iterable in which presence of elements from query_iterable is checked.
    all_present: bool
        If <True> return <True> only if all the elements from query_iterable are
        in reference_iterable.

    Returns
    -------
    bool
        <True> of <False> if all or any of the elements from one iterable are
        found in the another.
    """
    if all_present is True:
        return all([i in reference_iterable for i in query_iterable])
    else:
        return any([i in reference_iterable for i in query_iterable])


def remove_from_list(element,
                     iterable):
    """
    Return list without given element from that list. Conversely to built-in
    methods it is fruitful function.

    Parameters
    -------
    element: object
        Element to be removed from the list.
    iterable: list, tuple, set
        Iterable from which the element should be removed.

    Returns
    -------
    list
        Cleaned up from the element.

    """
    if element in list(iterable):
        iterable.remove(element)
    return iterable


def remove_char(string,
                iterable):
    """
    Return str without given elements from the iterable. More convenient than
    chaining the built-in replace methods.

    Parameters
    -------
    string: str
        String from which the characters from the iterable are removed.
    iterable: str, list, tuple, set
        Iterable with characters that are removed from the string.

    Returns
    -------
    str
        Without elements from the iterable.

    """
    for i in iterable:
        string = string.replace(i, "")
    return string


def all_possible_combinations_counter(subset_size,
                                      set_size):
    """
    Return a number (int) of all possible combinations of elements in size
    of a subset of a set.

    Parameters
    -------
    subset_size: int
        Size of the subset.
    set_size: int
        Size of the whole set.

    Returns
    -------
    int
        Number of all combinations.
    """
    f = math.factorial
    return f(set_size) / f(subset_size) / f(set_size - subset_size)



import pandas as _pd
import numpy as _np
from . import databases


def get_IDs_names(
    species,
):
    """
    Returns dict of KEGG Organism IDs as keys and biological names as values.

    Parameters
    -------
    species: list of str
        List of full biological names to convert into KEGG Organism IDs.

    Returns
    ------
    dict
    """
    kegg_db = databases.KEGG('Orthology')
    kegg_db.parse_organism_info(
        organism=None,
        reference_species=species,
        IDs=None,
        X_ref=None,
        KOs=None,
        IDs_only=True,
    )
    return {k.lower(): v for k, v in kegg_db.ID_name.items()}

def profilize_organism(
    organism,
    species,
    IDs=None,
    X_ref=None,
    KOs=None,
):
    """
    Returns pandas.DataFrame with Phylogenetic Profile for each ORF name of an
    organism.

    Parameters
    -------
    organism: str
        Full biological name of the organism.
    species: list of str
        List of full biological names to build the Phylogenetic Profile.
    IDs: str, path
        Filename of the KEGG Organism IDs. Downloaded to a temporary file if
        <None>.
    X_ref: str, path
        Filename of the ORF-KEGG Orthology Group cross-reference.
        Downloaded to a temporary file if <None>.
    KOs: str, path
        Filename of the KEGG Orthology Group-Organism cross-reference.
        Downloaded to a temporary file if <None>.

    Returns
    ------
    pandas.DataFrame
    """
    kegg_db = databases.KEGG('Orthology')
    kegg_db.parse_organism_info(
        organism=organism,
        reference_species=species,
        IDs=IDs,
        X_ref=X_ref,
        KOs=KOs,
    )
    return kegg_db.organism_info.drop(columns=databases.Columns.KEGG_ID)

def read_sga(
    filename,
    version=2,
):
    """
    Returns pandas.DataFrame with Genetic Interaction Network from
    the Costanzo's SGA experiment either version 1 or 2.

    Parameters
    -------
    filename: str, path
        Filename of the SGA.
    version: int
        Version number of the Costanzo's SGA experiment. 1 or 2 available.

    Returns
    -------
    pandas.DataFrame
    """
    if version == 1:
        sga = databases.SGA1()
    elif version == 2:
        sga = databases.SGA2()
    else:
        raise errors.ParserError("Only versions 1 and 2 of Costanzo's SGA experiment are supported.")
    sga.parse(filename=filename)
    return sga.sga

def merge_sga_profiles(
    sga,
    profiles
):
    """
    Returns Genetic Interaction Network from the Costanzo's SGA experiment with
    Phylogenetic Profiles.
    """
    merged = _pd.merge(
        left=sga,
        right=profiles,
        left_on=databases.Columns.ORF_Q,
        right_on=databases.Columns.ORF_ID,
        how="left",
    ).merge(
        right=profiles,
        left_on=databases.Columns.ORF_A,
        right_on=databases.Columns.ORF_ID,
        how="left",
        suffixes=(
            databases.Columns.QUERY_SUF,
            databases.Columns.ARRAY_SUF,
        )
    )
    merged.drop(
        columns=[
            databases.Columns.ORF_ID_Q,
            databases.Columns.ORF_ID_A,
        ],
        axis=1,
        inplace=True,
    )
    merged.dropna(inplace=True)
    merged.reset_index(
        drop=True,
        inplace=True,
    )
    return merged

def calculate_pss(
    network,
    method,
):
    """
    Returns Genetic Interaction Network from the Costanzo's SGA experiment with
    Profiles Similarity Score.

    Parameters
    -------
    network: pandas.DataFrame
    method: str

    Returns
    -------
    pandas.DataFrame
    """
    if method == 'pairwise':
        def pss(ar1, ar2):
            return sum(a == b for a, b in zip(ar1, ar2))
        pss_vect = _np.vectorize(pss)
        network[databases.Columns.PSS] = pss_vect(
            network[databases.Columns.PROF_Q].apply(lambda x: x.profile),
            network[databases.Columns.PROF_A].apply(lambda x: x.profile),
            )
    else:
        network[databases.Columns.PSS] = network.apply(
            lambda x: x[databases.Columns.PROF_Q].calculate_pss(
                x[databases.Columns.PROF_A],
                method=method,
                ),
            axis=1,
        )
    return network

def pss(ar1, ar2):
    return sum(a == b for a, b in zip(ar1, ar2))

pss_vect = _np.vectorize(pss)



from distutils.core import setup
from setuptools import find_packages
from prwlr import __version__
from prwlr import __author__


VERSION = __version__
AUTHOR = __author__

setup(
    name="prwlr",
    version=VERSION,
    author=AUTHOR,
    packages=find_packages(exclude=["*test*"]),
    install_requires=open("requirements.txt").readlines(),
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    long_description=open("README.md").read(),
    description="Annotate genetic interactions networks with phylogenetic profiles and other attributes.",
    author_email="dariusz.izak@ibb.waw.pl",
    url="https://github.com/dizak/prwlr",
    license="BSD",
    keywords=[
        'genetic',
        'interactions',
        'network',
        'phylogenetic',
        'profile',
        'module',
        'pathway',
        'kegg',
        'evolution',
    ],
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
    ],
)



# -*- coding: utf-8 -*-


from __future__ import print_function

class ParserError(Exception):
    """
    Inappropriate structure passed to parser.
    """
    pass


class ProfileError(Exception):
    """
    Different length of the reference and query.
    """
    pass


class SelectionFailWarning(Warning):
    """
    Dataframe boolean selection failure.
    """
    pass


class ExperimentalFeature(Warning):
    """
    This is an experimental feature. It may not work as expected nor work at
    all.
    """
    pass



# -*- coding: utf-8 -*-


from __future__ import print_function
import gc
import warnings
import pandas as pd
import math
from functools import partial
import numpy as np
import pathos.multiprocessing as ptmp
from tqdm import tqdm
from prwlr.errors import *
from prwlr.utils import *
from prwlr.databases import Columns as _DatabasesColumns
from prwlr.profiles import Profile as _Profile


class Columns(_DatabasesColumns):
    """
    Container for the columns names defined in this module.
    """
    P = "P"
    COUNT = "COUNT"
    COUNT_EXP = "COUNT_EXP"
    SCORE = "SCORE"
    SCORE_EXP = "SCORE_EXP"
    FOLD_CHNG = "FOLD_CHNG"
    SIM = "SIMILAR"
    DIS = "DISSIMILAR"
    MIR = "MIRROR"
    ITER = "ITERATION"
    DATAFRAME = "DATAFRAME"
    dtypes = {SIM: "uint32",
              DIS: "uint32",
              MIR: "uint32",
              ITER: "uint32",
              COUNT: "uint8",
              COUNT_EXP: "float32",
              SCORE: "float32",
              SCORE_EXP: "float32",
              FOLD_CHNG: "float32",
              P: "float32",
              _DatabasesColumns.PSS: _DatabasesColumns.dtypes[_DatabasesColumns.PSS]}


def _log_binomial_coeff(n,
                        k):
    # use multiplicative formula and calculate logarithms on the fly
    n = float(n)
    w = 0.0
    if k > n / 2:    # shorter loop
        k = int(n - k)
    for i in range(1, k + 1):
        w += math.log((n - i + 1.0) / float(i))
    return w

def _score(hit_num,
           prot_num,
           background_p):
    """
    Calculate logarithm of probability that given term was found hit_num times by chance.
    Use binomial distribution:
    log(P) = log((N  k)  * p**k * (1-p)**(N-k)) = log(N k) + k*log(p) + (N-k)*log(1-p)
    """
    prot_num = int(prot_num)
    hit_num = int(hit_num)
    log_p = (_log_binomial_coeff(prot_num, hit_num) +
             hit_num * math.log(background_p) +
             (prot_num - hit_num) * math.log(1.0 - background_p))
    return log_p

def calculate_enrichment(selected,
                         total,
                         col=Columns.PSS):
    """
    Returns enrichment table.

    Parameters
    -------
    selected: pandas.DataFrame
        Dataframe containing data of interest.
    total: pandas.DataFrame
        Dataframe containing all the samples.
    col: str
        Column name holding attribute to calculate enrichment on.

    Returns
    -------
        pandas.DataFrame
    Dataframe with enrichment scores and fold change.
    """
    if len(selected) == 0 or len(total) == 0:
        raise ValueError("selected and total dataframes must not be empty.")
    if len(selected) > len(total):
        raise ValueError("selected must not be longer bigger than total.")
    selected_bins = pd.DataFrame(selected.groupby(by=[col]).size(),
                                 columns=[Columns.COUNT]).reset_index()
    expected_bins = pd.DataFrame(total.groupby(by=[col]).size(),
                                 columns=[Columns.COUNT]).reset_index()
    selected_bins[Columns.P] = expected_bins[Columns.COUNT].apply(lambda x: float(x) / float(len(total)))
    selected_bins[Columns.COUNT_EXP] = selected_bins.apply(lambda x: len(selected) * x[Columns.P],
                                                        axis=1)
    selected_bins[Columns.SCORE] = selected_bins.apply(lambda x: Columns._score(x[Columns.COUNT],
                                                                          len(selected),
                                                                          x[Columns.P]),
                                                    axis=1)
    selected_bins[Columns.SCORE_EXP] = selected_bins.apply(lambda x: Columns._score(x[Columns.COUNT_EXP],
                                                                              len(selected),
                                                                              x[Columns.P]),
                                                        axis=1)
    selected_bins[Columns.FOLD_CHNG] = np.log2(selected_bins[Columns.COUNT] /
                                            selected_bins[Columns.COUNT_EXP])
    selected_bins = selected_bins.astype({k: v for k, v in Columns.dtypes.items()
                                         if k in selected_bins.columns})
    return selected_bins

def binomial_pss_test(desired_pss,
                      selected,
                      total,
                      test_size=1000):
    """
    Draws samples of desired PSS from binomial distribution. Uses
    numpy.random.binomial.

    Parameters
    -------
    desired_pss: int
        Profiles Similarity Score of interest. Interactions with this or
        higher values are considered success.
    selected: pandas.DataFrame
        Dataframe containing data of interest. Interactions with this
        values are subject of the test.
    total: pandas.DataFrame
        Dataframe containing all the samples.

    Returns
    -------
    dict with int values
        <complete> - number of interactions with PSS higher than the
        test single results.
        <average> - average of single test values.
    """
    p = float(len(total[total[Columns.PSS] >= desired_pss])) / float(len(total))
    n = float(len(selected))
    real_val = len(selected[selected[Columns.PSS] >= desired_pss])
    test = np.random.binomial(n, p, test_size)
    return {"complete": sum(test <= real_val),
            "average": sum(test) / len(test)}
